/**
  Ambient Occlusion shader
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2012-04-11 (Public Domain)
*/
#include "physics/world.h" /* physics::library and physics::object */
#include "ogl/glsl.h"
#include <fstream>
#include <sstream>

/* Just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"

/* Current camera position, for use in raytracers */
vec3 rayObjectCamera;
vec3 lightLocation;
/* Current 4x4 shadow projection matrix */
float shadowMat[16];

// Check GL for errors after the last command
void oglCheck(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	printf("Error %s at %s\n",where, (const char *)errString);
	abort();
}

float randFloat(void) {
	return (rand()&0xffff)*(1.0/(float)0xffff);
}
vec3 randVec(void) {
	return vec3(randFloat(),randFloat(),randFloat());
}


/******************** 3D Mesh Object **********************/
class tri {
public:
	int verts[3];
};

class meshObject : public physics::object {
	std::vector<vec3> verts; // 3D positions
	std::vector<tri> tris; // vertex numbers
	
	std::vector<vec3> movedVerts;
	std::vector<vec3> normals;
public:
	meshObject(const char *objFileName) 
		:physics::object(0.01)  /* <- our timestep, in seconds */
	{
		std::ifstream file("cat.obj");
		std::string line;
		while (std::getline(file,line)) {
			if (line[0]=='v') { /* vertex */
				vec3 v;
				std::stringstream ss(line);
				char c; ss>>c; // skip the 'c'
				ss>>v.x>>v.y>>v.z;

				/* OBJ format has Y up, we want Z up */
				std::swap(v.y,v.z);
				v.y=-v.y;

				/* push model up vertically */
				v.z+=0.2;
				if (!ss.fail())
					verts.push_back(v);
				else {
					std::cout<<"Error in .obj v line!  What is "<<line<<"!?\n";
				}
			}
			else if (line[0]=='f') { /* face: FIXME assumes triangle */
				tri t;
				std::stringstream ss(line);
				char c; ss>>c; // skip the 'f'
				for (unsigned int j=0;j<3;j++)
				{
					ss>>t.verts[j];
					t.verts[j]--; // from 1-based obj file to 0-based truth!
				}
				if (!ss.fail())
					tris.push_back(t);
				else {
					std::cout<<"Error in .obj f line!  What is "<<line<<"!?\n";
				}
			}
		}
		movedVerts.resize(verts.size());
		normals.resize(verts.size());
	}
	
	void simulate(physics::library &lib) { }
	
	void draw(physics::library &lib) {
		
		// Move the vertices
		for (unsigned int i=0;i<verts.size();i++)
			movedVerts[i]=verts[i]+ // old point
					0.1*sin(lib.time)* // how far
						vec3(
							sin(4.0*verts[i].y), // x shift
							0.0, 
							cos(5.0*verts[i].x)); // z shift
		
		// Calculate new normals 
		for (unsigned int i=0;i<verts.size();i++)
			normals[i]=vec3(0.0); // zero out the normal accumulator
		
		for (unsigned int tri=0;tri<tris.size();tri++) 
		{
			vec3 v0=movedVerts[tris[tri].verts[0]];
			vec3 v1=movedVerts[tris[tri].verts[1]];
			vec3 v2=movedVerts[tris[tri].verts[2]];
			vec3 triNormal=normalize(cross(v1-v0,v2-v0));
			for (unsigned int j=0;j<3;j++) {
				normals[tris[tri].verts[j]]+=triNormal;
			}
		}


		/* Load up shader */
		static GLhandleARB triprog=MAKE_PROGRAM_OBJECT3(
		/* vertex program */
			void main(void) {
			   myColor = gl_Color;
			   myNormal = normalize(gl_Normal);
			   /* Include everything in onscreen position */
			   gl_Position=gl_ModelViewProjectionMatrix*gl_Vertex;
			   /* Don't project, but do modelview transforms */
			   location=(vec3)(gl_ModelViewMatrix*gl_Vertex);
			}
		,
		/* shared vertex/fragment code */
			varying vec4 myColor;
			varying vec3 location; // world coordinates of our pixel
			varying vec3 myNormal; 

		,
		/* fragment program */
			uniform sampler2D shadowTex[1];
			uniform mat4 shadowMat; // takes world coords to shadow coords (perspective)
			uniform vec3 camera; // world coordinates of camera
			uniform vec3 lightLocation; // world coordinates of light
			
			/* Return 1.0 if this point is in light, 0.0 if it is fully shadowed. */
			float in_light(vec3 loc) {
				vec4 my4=shadowMat*vec4(loc,1.0); /* run shadow location through projection matrix */
				vec3 my=vec3(my4/my4.w)*0.5+0.5; /* perspective divide, and [-1,+1] to [0,1] */
				float del=(1.0/1024); /* distance in texture coordinates to move sample */
				if (my.z<del || my.z>1.0-del || my.x<del || my.x>1.0-del || my.y<del || my.y>1.0-del)
					return 1.0; // default to lit (if out of bounds)
				
				float shadowness=0.0;
				for (float dy=-1.0;dy<=1.0;dy+=1.0)
				for (float dx=-1.0;dx<=1.0;dx+=1.0)
				{
					vec4 st=texture2D(shadowTex[0],vec2(my)+(1.0/1024.0)*vec2(dx,dy));
					float highestZ=st.x;
					// FIXME: should move my.z in plane based on my normal here...
					if (my.z>highestZ+3.0/1000) shadowness+=(1.0/9.0);
				}
				return 1.0-shadowness;
			}
			
			void main(void) {
				vec3 lightDir=normalize(lightLocation-location);
				vec3 N=normalize(myNormal);
				float ambient=0.1;
				float diffuse=dot(lightDir,N);
				if (diffuse<0.0) diffuse=-0.4*diffuse;
				
				
				// gl_FragColor=vec4(vec3(N.z,diffuse,in_light(location)),1.0);
				gl_FragColor=vec4(vec3(diffuse*in_light(location)),1.0);
			}
		);
		
		glUseProgramObjectARB(triprog);
		int ul=glGetUniformLocationARB(triprog,"shadowTex");
		int shadowUnits[1]={7};
		glUniform1ivARB(ul,1,shadowUnits); /* shadow texture on unit 7 */
		ul=glGetUniformLocationARB(triprog,"shadowMat");
		glUniformMatrix4fv(ul,1,0,&shadowMat[0]);
		ul=glGetUniformLocationARB(triprog,"camera");
		glUniform3fvARB(ul,1,rayObjectCamera);
		ul=glGetUniformLocationARB(triprog,"lightLocation");
		glUniform3fvARB(ul,1,lightLocation);
		
		/* Draw the triangles */
		glColor3f(1.0,1.0,1.0);
		glBegin(GL_TRIANGLES);
		
		/* Draw mesh */
		for (unsigned int i=0;i<tris.size();i++)
		{
			for (unsigned int j=0;j<3;j++)
			{
				int theVertex=tris[i].verts[j];
				vec3 v=movedVerts[theVertex];
				vec3 normal=normals[theVertex];
				
				glNormal3fv(normalize(normal));
				
				glVertex3fv(v);
			}
		}
		
		/* Draw ground */
		float z=-1.0+0.001;
		float sz=30.0;
		glNormal3f(0,0,1);
		glVertex3f(-sz,-sz,z); glVertex3f(+sz,-sz,z); glVertex3f(+sz,+sz,z); 
		glVertex3f(-sz,-sz,z); glVertex3f(-sz,+sz,z); glVertex3f(+sz,+sz,z); 
		
		glEnd();
		glUseProgramObjectARB(0);
	}
};


/* Called to create a new simulation */
void physics_setup(physics::library &lib) {
	lib.world->add(new meshObject("cat.obj"));
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
		oglCheck("before drawing");
	/* Point and line smoothing needs usual alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	
	int w=1024,h=1024;
	static GLuint texDepth=0;
	if (texDepth==0) {
		oglCheck("before texture work");
		glGenTextures(1,&texDepth);
		glBindTexture(GL_TEXTURE_2D,texDepth);
		glTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT24, 
				w,h, 0, GL_DEPTH_COMPONENT,GL_FLOAT,0);
		glGenerateMipmap(GL_TEXTURE_2D); //<- slow for depth textures?
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	/*
		glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_COMPARE_MODE_ARB,
			GL_COMPARE_R_TO_TEXTURE_ARB); // magic percent-closer-filtering mode (otherwise, you get black!)
		glTexParameteri(GL_TEXTURE_2D, 
			GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
	*/
		oglCheck("after texture work");
	}
	
	static GLuint fbo=0;
	if (fbo==0) {
		glGenFramebuffers(1,&fbo);
		oglCheck("before FBO setup");
		glBindFramebuffer(GL_FRAMEBUFFER,fbo);
		glFramebufferTexture2D(GL_FRAMEBUFFER,
			GL_DEPTH_ATTACHMENT,GL_TEXTURE_2D,texDepth,0);
		glBindFramebuffer(GL_FRAMEBUFFER,0);
		oglCheck("after FBO setup");
	}
	
	oglCheck("before FBO frame");
	glBindFramebuffer(GL_FRAMEBUFFER,fbo);
	glViewport(0,0,w,h);
	glClearColor(1,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	oglCheck("after FBO setup");
	
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	float scale=0.4;
	//glScalef(scale,scale,scale); /* negative on Z, since +z means farther away */
	gluPerspective(50.0,w/(float)h,9.1,15.1);
	float lAngle=lib.time*(1.0/5.0)*3.1415;
	float lRadius=5.0;
	// Light source is circulating around origin
	lightLocation=vec3(lRadius*cos(lAngle),lRadius*sin(lAngle),10.0);
	rayObjectCamera=lightLocation;
	gluLookAt(VEC3_TO_XYZ(rayObjectCamera), 
		0.0,0.0,0.0, // center of view 
		0.0,1.0,0.0 // up vector 
	);
	glGetFloatv(GL_PROJECTION_MATRIX,&shadowMat[0]); /* read back matrix */
	glDisable(GL_TEXTURE_2D);
	

	/* Do depth while drawing objects */
	lib.world->draw(lib);


	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
	oglCheck("after FBO draw");
	glBindFramebuffer(GL_FRAMEBUFFER,0); /* back to onscreen */
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));

	glActiveTexture(GL_TEXTURE7);
	glBindTexture(GL_TEXTURE_2D,texDepth);
	glEnable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE0);
	
	rayObjectCamera=camera;
	/* Do depth while drawing objects */
	glEnable(GL_DEPTH_TEST);
	lib.world->draw(lib);
	
	oglCheck("after drawing");
}


